use std::sync::Arc;

use dagger_sdk::{ContainerPublishOptsBuilder, QueryContainerOptsBuilder};
use eyre::Context;
use tokio::sync::mpsc;

#[tokio::main]
async fn main() -> eyre::Result<()> {
    color_eyre::install().unwrap();

    let client = dagger_sdk::connect().await?;

    let (tx, mut rx) = mpsc::channel(2);

    async_scoped::TokioScope::scope_and_block(|s| {
        for platform in &[("linux/amd64", "amd64"), ("linux/arm64", "aarch64")] {
            let client = client.clone();
            let tx = tx.clone();
            let f = || async move {
                let container = run_wasmer(client, platform.0, platform.1)
                    .await
                    .context("failed to run wasmer")
                    .unwrap();
                let container_id = container
                    .id()
                    .await
                    .context("failed to get container id")
                    .unwrap();

                println!("done building for platform: {}", platform.0);

                tx.send(container_id)
                    .await
                    .context("failed to send container id")
                    .unwrap();
            };

            s.spawn(f());
        }
    });

    drop(tx);
    let mut containers: Vec<dagger_sdk::ContainerId> = Default::default();
    while let Some(container) = rx.recv().await {
        containers.push(container);
    }

    println!("publishing image!");

    client
        .container()
        .publish_opts(
            "git.front.kjuulh.io/kjuulh/wasmer:debian",
            ContainerPublishOptsBuilder::default()
                .platform_variants(containers)
                .build()?,
        )
        .await?;

    Ok(())
}

async fn run_wasmer(
    client: Arc<dagger_sdk::Query>,
    platform: &str,
    architecture: &str,
) -> eyre::Result<dagger_sdk::Container> {
    let wasmer_version = "v3.3.0";
    let wasmer_architecture = architecture;
    let wasmer_download_url =
        format!("https://github.com/wasmerio/wasmer/releases/download/{wasmer_version}/wasmer-linux-{wasmer_architecture}.tar.gz");

    let debian = build_debian(client, platform).await?;

    let wasmer = debian
        .with_exec(vec!["wget", &wasmer_download_url, "-O", "wasmer.tar.gz"])
        .with_exec(vec!["tar", "-xvf", "wasmer.tar.gz"])
        .with_exec(vec!["mv", "bin/wasmer", "/usr/bin/wasmer"])
        .with_exec(vec!["wasmer", "-v"]);

    let exit = wasmer.exit_code().await?;
    if exit != 0 {
        eyre::bail!("failed with exitcode: {exit}");
    }

    Ok(wasmer)
}

async fn build_debian(
    client: Arc<dagger_sdk::Query>,
    platform: &str,
) -> eyre::Result<dagger_sdk::Container> {
    let debian = client
        .container_opts(
            QueryContainerOptsBuilder::default()
                .platform(platform)
                .build()?,
        )
        .from("debian:bookworm")
        .with_workdir("/tmp/")
        .with_exec(vec!["apt", "update", "-y"])
        .with_exec(vec!["apt", "install", "wget", "xz-utils", "-y"]);

    Ok(debian)
}
